/* -*- mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/*
 Copyright (C) 2016 Klaus Spanderen

 This file is part of QuantLib, a free-software/open-source library
 for financial quantitative analysts and developers - http://quantlib.org/

 QuantLib is free software: you can redistribute it and/or modify it
 under the terms of the QuantLib license.  You should have received a
 copy of the license along with this program; if not, please email
 <quantlib-dev@lists.sf.net>. The license is also available online at
 <http://quantlib.org/license.shtml>.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.
*/


#include "utilities.hpp"
#include "squarerootclvmodel.hpp"

#include <ql/quotes/simplequote.hpp>
#include <ql/time/daycounters/actualactual.hpp>
#include <ql/time/daycounters/actual365fixed.hpp>
#include <ql/time/calendars/nullcalendar.hpp>
#include <ql/instruments/impliedvolatility.hpp>
#include <ql/instruments/forwardvanillaoption.hpp>
#include <ql/math/statistics/statistics.hpp>
#include <ql/math/integrals/gausslobattointegral.hpp>
#include <ql/math/randomnumbers/rngtraits.hpp>
#include <ql/math/randomnumbers/sobolbrownianbridgersg.hpp>
#include <ql/processes/squarerootprocess.hpp>
#include <ql/pricingengines/blackcalculator.hpp>
#include <ql/pricingengines/vanilla/analyticeuropeanengine.hpp>
#include <ql/pricingengines/forward/forwardengine.hpp>
#include <ql/methods/montecarlo/pathgenerator.hpp>
#include <ql/termstructures/volatility/equityfx/hestonblackvolsurface.hpp>

#include <ql/experimental/models/squarerootclvmodel.hpp>
#include <ql/experimental/volatility/sabrvoltermstructure.hpp>

#if defined(__GNUC__) && (((__GNUC__ == 4) && (__GNUC_MINOR__ >= 8)) || (__GNUC__ > 4))
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-local-typedefs"
#endif
#include <boost/bind.hpp>
#if defined(__GNUC__) && (((__GNUC__ == 4) && (__GNUC_MINOR__ >= 8)) || (__GNUC__ > 4))
#pragma GCC diagnostic pop
#endif

#include <boost/make_shared.hpp>
#include <boost/assign/std/vector.hpp>
#include <boost/math/distributions/non_central_chi_squared.hpp>

#include <iostream>


using namespace QuantLib;
using namespace boost::assign;
using namespace boost::unit_test_framework;


namespace {
    class CLVModelPayoff : public PlainVanillaPayoff {
      public:
        CLVModelPayoff(Option::Type type, Real strike,
                             const boost::function<Real(Real)> g)
        : PlainVanillaPayoff(type, strike),
          g_(g) { }

        Real operator()(Real x) const {
            return PlainVanillaPayoff::operator()(g_(x));
        }

      private:
        const boost::function<Real(Real)> g_;
    };
}


void SquareRootCLVModelTest::testSquareRootCLVVanillaPricing() {
    BOOST_TEST_MESSAGE(
        "Testing vanilla option pricing with square root kernel process...");

    SavedSettings backup;

    const Date todaysDate(5, Oct, 2016);
    Settings::instance().evaluationDate() = todaysDate;

    const DayCounter dc = ActualActual();
    const Date maturityDate = todaysDate + Period(3, Months);
    const Time maturity = dc.yearFraction(todaysDate, maturityDate);

    const Real s0 = 100;
    const Handle<Quote> spot(boost::make_shared<SimpleQuote>(s0));

    const Rate r = 0.08;
    const Rate q = 0.03;
    const Volatility vol = 0.3;

    const Handle<YieldTermStructure> rTS(flatRate(r, dc));
    const Handle<YieldTermStructure> qTS(flatRate(q, dc));
    const Handle<BlackVolTermStructure> volTS(flatVol(todaysDate, vol, dc));
    const Real fwd = s0*qTS->discount(maturity)/rTS->discount(maturity);

    const boost::shared_ptr<GeneralizedBlackScholesProcess> bsProcess(
        boost::make_shared<GeneralizedBlackScholesProcess>(
            spot, qTS, rTS, volTS));

    const Real kappa       = 1.0;
    const Real theta       = 0.06;
    const Volatility sigma = 0.2;
    const Real x0          = 0.09;

    const boost::shared_ptr<SquareRootProcess> sqrtProcess(
        boost::make_shared<SquareRootProcess>(theta, kappa, sigma, x0));

    const std::vector<Date> maturityDates(1, maturityDate);

    const SquareRootCLVModel model(
        bsProcess, sqrtProcess, maturityDates, 18, 1-1e-14, 1e-14);

    const Array x = model.collocationPointsX(maturityDate);
    const Array y = model.collocationPointsY(maturityDate);

    const LagrangeInterpolation g(x.begin(), x.end(), y.begin());

    const Real df  = 4*theta*kappa/(sigma*sigma);
    const Real ncp = 4*kappa*std::exp(-kappa*maturity)
            / (sigma*sigma*(1-std::exp(-kappa*maturity)))*sqrtProcess->x0();

    typedef boost::math::non_central_chi_squared_distribution<Real>
        chi_squared_type;

    const chi_squared_type dist(df, ncp);

    const Real strikes[] = { 50, 75, 100, 125, 150, 200 };
    for (Size i=0; i < LENGTH(strikes); ++i) {
        const Real strike = strikes[i];
        const Option::Type optionType =
            (strike > fwd) ? Option::Call : Option::Put;

        const Real expected = BlackCalculator(
            optionType, strike, fwd,
            std::sqrt(volTS->blackVariance(maturity, strike)),
            rTS->discount(maturity)).value();

        const CLVModelPayoff clvModelPayoff(optionType, strike, g);

        const boost::function<Real(Real)> f =
            boost::bind(std::multiplies<Real>(),
                boost::bind(&CLVModelPayoff::operator(), &clvModelPayoff, _1),
                boost::bind<Real>(boost::math::pdf<chi_squared_type, Real>,
                    dist, _1) );

        const Real calculated = GaussLobattoIntegral(1000, 1e-6)(
            f, x.front(), x.back()) * rTS->discount(maturity);

        const Real tol = 1e-3;
        if (std::fabs(expected - calculated) > tol) {
            BOOST_FAIL("failed to reproduce option SquaredCLVMOdel prices"
                    << "\n    time:       " << maturityDate
                    << "\n    strike:     " << strike
                    << "\n    expected:   " << expected
                    << "\n    calculated: " << calculated);
        }
    }
}

void SquareRootCLVModelTest::testSquareRootCLVMappingFunction() {
    BOOST_TEST_MESSAGE(
        "Testing mapping function of the square root kernel process...");

    SavedSettings backup;

    const Date todaysDate(16, Oct, 2016);
    Settings::instance().evaluationDate() = todaysDate;
    const Date maturityDate = todaysDate + Period(1, Years);

    const DayCounter dc = Actual365Fixed();

    const Real s0 = 100;
    const Handle<Quote> spot(boost::make_shared<SimpleQuote>(s0));

    const Rate r = 0.05;
    const Rate q = 0.02;

    const Handle<YieldTermStructure> rTS(flatRate(r, dc));
    const Handle<YieldTermStructure> qTS(flatRate(q, dc));

    //SABR
    const Real beta =  0.95;
    const Real alpha=  0.2;
    const Real rho  = -0.9;
    const Real gamma=  0.8;

    const Handle<BlackVolTermStructure> sabrVol(
        boost::make_shared<SABRVolTermStructure>(
            alpha, beta, gamma, rho, s0, r, todaysDate, dc));

    const boost::shared_ptr<GeneralizedBlackScholesProcess> bsProcess(
        boost::make_shared<GeneralizedBlackScholesProcess>(
            spot, qTS, rTS, sabrVol));

    std::vector<Date> calibrationDates(1, todaysDate + Period(1, Weeks));
    calibrationDates.reserve(daysBetween(todaysDate, maturityDate)/7 + 1);
    while (calibrationDates.back() < maturityDate)
        calibrationDates.push_back(calibrationDates.back() + Period(1, Weeks));

    // sqrt process
    const Real kappa       = 1.0;
    const Real theta       = 0.09;
    const Volatility sigma = 0.2;
    const Real x0          = 0.09;

    const boost::shared_ptr<SquareRootProcess> sqrtProcess(
        boost::make_shared<SquareRootProcess>(theta, kappa, sigma, x0));

    const SquareRootCLVModel model(
        bsProcess, sqrtProcess, calibrationDates, 18, 1-1e-14, 1e-14);

    const boost::function<Real(Time, Real)> g = model.g();

    const Real strikes[] = { 80, 100, 120 };
    const Size offsets[] = { 7, 14, 28, 91, 182, 183, 184, 185, 186, 365 };
    for (Size i=0; i < LENGTH(offsets); ++i) {
        const Date m = todaysDate + Period(offsets[i], Days);
        const Time t = dc.yearFraction(todaysDate, m);

        const Real df  = 4*theta*kappa/(sigma*sigma);
        const Real ncp = 4*kappa*std::exp(-kappa*t)
                / (sigma*sigma*(1-std::exp(-kappa*t)))*sqrtProcess->x0();

        typedef boost::math::non_central_chi_squared_distribution<Real>
            chi_squared_type;

        const chi_squared_type dist(df, ncp);

        const Real fwd = s0*qTS->discount(m)/rTS->discount(m);

        for (Size j=0; j < LENGTH(strikes); ++j) {
            const Real strike = strikes[j];
            const Option::Type optionType =
                (strike > fwd) ? Option::Call : Option::Put;

            const Real expected = BlackCalculator(
                optionType, strike, fwd,
                std::sqrt(sabrVol->blackVariance(m, strike)),
                rTS->discount(m)).value();

            const CLVModelPayoff clvModelPayoff(
                optionType, strike, boost::bind(g, t, _1));

            const boost::function<Real(Real)> f =
                boost::bind(std::multiplies<Real>(),
                    boost::bind(
                        &CLVModelPayoff::operator(), &clvModelPayoff, _1),
                    boost::bind<Real>(boost::math::pdf<chi_squared_type, Real>,
                        dist, _1) );

            const Array x = model.collocationPointsX(m);
            const Real calculated = GaussLobattoIntegral(1000, 1e-3)(
                f, x.front(), x.back()) * rTS->discount(m);

            const Real tol = 1.5e-2;
            if (std::fabs(calculated - expected) > tol) {
                BOOST_FAIL("failed to reproduce option SquaredCLVMOdel prices"
                        << "\n    time:       " << m
                        << "\n    strike:     " << strike
                        << "\n    expected:   " << expected
                        << "\n    calculated: " << calculated);
            }
        }
    }
}

void SquareRootCLVModelTest::testForwardSkew() {
    BOOST_TEST_MESSAGE(
        "Testing forward skew dynamics with square root kernel process...");

    SavedSettings backup;

    const Date todaysDate(16, Oct, 2016);
    Settings::instance().evaluationDate() = todaysDate;
    const Date endDate = todaysDate + Period(5, Years);

    const DayCounter dc = Actual365Fixed();

    // Heston model is used to generate an arbitrage free volatility surface
    const Real s0    =  100;
    const Real r     =  0.1;
    const Real q     =  0.05;
    const Real v0    =  0.09;
    const Real kappa =  1.0;
    const Real theta =  0.06;
    const Real sigma =  0.4;
    const Real rho   = -0.75;

    const Handle<Quote> spot(boost::make_shared<SimpleQuote>(s0));
    const Handle<YieldTermStructure> rTS(flatRate(r, dc));
    const Handle<YieldTermStructure> qTS(flatRate(q, dc));

    const Handle<HestonModel> hestonModel(
        boost::make_shared<HestonModel>(
            boost::make_shared<HestonProcess>(
                rTS, qTS, spot, v0, kappa, theta, sigma, rho)));

    const Handle<BlackVolTermStructure> hestonVolTS(
        boost::make_shared<HestonBlackVolSurface>(hestonModel));

    const boost::shared_ptr<GeneralizedBlackScholesProcess> bsProcess(
        boost::make_shared<GeneralizedBlackScholesProcess>(
            spot, qTS, rTS, hestonVolTS));

    const Real sKappa = 0.025;
    const Real sTheta = 2.0;
    const Real sSigma = 0.2;
    const Real sX0    = 2.5;

    const boost::shared_ptr<SquareRootProcess> sqrtProcess(
        boost::make_shared<SquareRootProcess>(
            sTheta, sKappa, sSigma, sX0));

    typedef boost::math::non_central_chi_squared_distribution<Real>
        chi_squared_type;

    std::vector<Date> calibrationDates(1, todaysDate + Period(3, Months));
    while (calibrationDates.back() < endDate)
        calibrationDates.push_back(calibrationDates.back() + Period(3, Months));

    const SquareRootCLVModel clvSqrtModel(
        bsProcess, sqrtProcess, calibrationDates, 18, 1-1e-14, 1e-14);

    const boost::function<Real(Time, Real)> gSqrt = clvSqrtModel.g();

    for (Size i=0, n=calibrationDates.size()-2; i < n; ++i) {
        const Date resetDate = calibrationDates[i];
        const Date maturityDate = calibrationDates[i+2];

        const Time t0 = dc.yearFraction(todaysDate, resetDate);
        const Time t1 = dc.yearFraction(todaysDate, maturityDate);

        const Real df  = 4*sTheta*sKappa/(sSigma*sSigma);
        const Real ncp = 4*sKappa*std::exp(-sKappa*t0)
            / (sSigma*sSigma*(1-std::exp(-sKappa*t0)))*sX0;

        const chi_squared_type dist(df, ncp);

        const Real ncp1 = 4*sKappa*std::exp(-sKappa*(t1-t0))
            / (sSigma*sSigma*(1-std::exp(-sKappa*(t1-t0))));

        const LowDiscrepancy::ursg_type ursg
            = LowDiscrepancy::ursg_type(2, 1235ul);

        const Real strikes[] = {
            0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.1, 1.2,
            1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2.0
        };

        std::vector<GeneralStatistics> stats(LENGTH(strikes));

        const Size nScenarios = 10000;
        for (Size j=0; j < nScenarios; ++j) {
            const std::vector<Real>& path = ursg.nextSequence().value;

            const Real x1 = boost::math::quantile(dist, path[0]);
            const Real u1 =
                sSigma*sSigma*(1-std::exp(-sKappa*t0))/(4*sKappa) * x1;

            const Real x2 = boost::math::quantile(
                chi_squared_type(df, ncp1*u1), path[1]);
            const Real u2 =
                sSigma*sSigma*(1-std::exp(-sKappa*(t1-t0)))/(4*sKappa) * x2;
            const Real X2 =
                u2*4*sKappa/(sSigma*sSigma*(1-std::exp(-sKappa*t1)));

            const Real s1 = gSqrt(t0, x1);
            const Real s2 = gSqrt(t1, X2);

            for (Size k=0; k < LENGTH(strikes); ++k) {
                const Real strike = strikes[k];

                const Real payoff = (strike < 1.0)
                    ?  s1 * std::max(0.0, strike - s2/s1)
                    :  s1 * std::max(0.0, s2/s1 - strike);

                stats[k].add(payoff);
            }
        }

        const boost::shared_ptr<SimpleQuote> vol(
            boost::make_shared<SimpleQuote>(0.1));
        const Handle<BlackVolTermStructure> volTS(flatVol(todaysDate, vol, dc));

        const boost::shared_ptr<GeneralizedBlackScholesProcess> bsProcess(
            new GeneralizedBlackScholesProcess(spot, qTS, rTS, volTS));

        const boost::shared_ptr<PricingEngine> fwdEngine(
            boost::make_shared<ForwardVanillaEngine<AnalyticEuropeanEngine> >(
                bsProcess));

        const boost::shared_ptr<Exercise> exercise(
            boost::make_shared<EuropeanExercise>(maturityDate));

        for (Size k=0; k < LENGTH(strikes); ++k) {
            const Real strike = strikes[k];
            const Real npv = stats[k].mean() * rTS->discount(maturityDate);

            const boost::shared_ptr<StrikedTypePayoff> payoff(
                boost::make_shared<PlainVanillaPayoff>(
                    (strike < 1.0) ? Option::Put : Option::Call, strike));

            const boost::shared_ptr<ForwardVanillaOption> fwdOption(
                boost::make_shared<ForwardVanillaOption>(
                    strike, resetDate, payoff, exercise));

            const Volatility implVol =
                detail::ImpliedVolatilityHelper::calculate(
                    *fwdOption, *fwdEngine, *vol, npv, 1e-8, 200, 1e-4, 2.0);

            std::cout << implVol << ", ";
//            std::cout << dc.yearFraction(todaysDate, resetDate)
//                    << " " << strike << " " << implVol << std::endl;
        }

    }


    std::cout /*<< calibrationDates.size()*/ << std::endl;

}

test_suite* SquareRootCLVModelTest::experimental() {
    test_suite* suite = BOOST_TEST_SUITE("NormalCLVModel tests");

    suite->add(QUANTLIB_TEST_CASE(
        &SquareRootCLVModelTest::testSquareRootCLVVanillaPricing));
    suite->add(QUANTLIB_TEST_CASE(
        &SquareRootCLVModelTest::testSquareRootCLVMappingFunction));
    suite->add(QUANTLIB_TEST_CASE(
        &SquareRootCLVModelTest::testForwardSkew));

    return suite;
}
